from collections.abc import Collection
from pathlib import Path

from rich.columns import Columns
from rich.text import Text

from cx_studio.utils import PathUtils, FunctionalUtils
from ..appenv import appenv
from .preset import Preset


class InputScanner:
    def __init__(self, inputs: Collection[str | Path]):
        self._inputs: list[str | Path] = list(inputs)
        self._task_id = appenv.progress.add_task("预处理输入项…", visible=False)

    def __enter__(self):
        appenv.progress.update(self._task_id, visible=True)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        appenv.progress.stop_task(self._task_id)
        appenv.progress.update(self._task_id, visible=False)

        return False

    def __del__(self):
        appenv.progress.remove_task(self._task_id)

    @staticmethod
    def is_preset(path: str | Path) -> bool:
        path = Path(path)
        suffix = path.suffix.lower()
        if suffix == ".toml":
            return True
        if suffix == "":
            p_path = PathUtils.force_suffix(path, ".toml")
            return p_path.exists()
        return False

    def add_inputs(self, *paths) -> "InputScanner":
        for p in FunctionalUtils.flatten_list(*paths):
            self._inputs.append(p)
        return self

    @staticmethod
    def _print_result(source: str | Path, is_preset: bool):
        result = (
            Text("配置文件路径", style="cyan", justify="right")
            if is_preset
            else Text("媒体来源路径", style="green", justify="right")
        )
        path = Text(str(source), style="yellow", justify="left")
        appenv.whisper(Columns([path, result], expand=True))

    def scan(self) -> tuple[list[Preset], list[Path]]:
        presets: list[Preset] = []
        sources: list[Path] = []

        appenv.whisper("检索待处理路径并从中解析配置文件...")

        for input_path in appenv.progress.track(self._inputs, task_id=self._task_id):
            if self.is_preset(input_path):
                preset = Preset.load(input_path)
                presets.append(preset)
                self._print_result(input_path, True)
            else:
                sources.append(Path(input_path))
                self._print_result(input_path, False)

            # time.sleep(0.5)

        return presets, sources



import importlib.resources
from cx_studio.utils import TextUtils
from cx_wealth import WealthHelp
from cx_wealth import rich_types as r


class MKHelp(WealthHelp):
    def __init__(self):
        super().__init__(prog="mediakiller")

        trans_opts = self.add_group("转码选项", "控制转码操作的选项")
        trans_opts.add_action(
            "inputs",
            nargs="+",
            metavar="FILE",
            description="指定（多个）需要处理的文件，包括[u]预设文件[/]和[u]源文件路径[/]。",
        )
        trans_opts.add_action(
            "-o",
            "--output",
            metavar="DIR",
            description="指定目标文件夹，如果不指定则默认为当前工作目录。",
        )
        trans_opts.add_action(
            "--sort",
            metavar="source|preset|target|x",
            description=TextUtils.unwrap(
                """设置任务的排序模式，在执行任务之前将会按照指定的模式进行排序。
            四种模式分别为[u]按源文件路径排序[/]、[u]按预设排序[/]、[u]按目标文件路径排序[/]、[u]按输入顺序排序[/]。"""
            ),
        )

        trans_opts.add_action(
            "-j",
            "--jobs" "--max-workers",
            metavar="NUM",
            description=TextUtils.unwrap(
                """
                设置并行工作进程的数量，默认为 1 。
                不建议设置大于 2 的数值，除非你知道你在干什么。
                """
            ),
        )

        trans_opts.add_action(
            "-y",
            "--overwrite",
            description="启用[red bold]强制覆盖模式[/]后，将忽略配置文件中定义的覆盖选项。",
        )
        trans_opts.add_action(
            "-n",
            "--no-overwrite",
            description="启用[green bold]安全模式[/]之后，将忽略一切覆盖选项，无论如何也不会覆盖目标文件。",
        )

        basic_opts = self.add_group("其它操作", "运行除了转码之外的操作")
        basic_opts.add_action(
            "-g",
            "--generate",
            metavar="PRESET",
            description="以示例内容生成预设文件。[uu]示例文件不可直接运行！[/]",
            nargs="+",
        )
        basic_opts.add_action(
            "-s",
            "--save",
            metavar="FILE",
            description="保存转码任务为脚本文件，将不再主动进行转码操作。",
        )
        basic_opts.add_action(
            "-c",
            "--continue",
            description="加载上次的[u]所有[/]转码任务并重新执行。\n如果想要继续运行未完成的任务，建议附加 -n 选项。",
        )

        misc_opts = self.add_group("杂项")
        misc_opts.add_action("-h", "--help", description="显示此帮助信息")
        misc_opts.add_action(
            "--tutorial", "--full-help", description="显示完整的教程内容"
        )
        misc_opts.add_action(
            "-d", "--debug", description="开启调试模式以观察更多的后台信息"
        )
        misc_opts.add_action(
            "-p",
            "--pretend",
            description="启用[bold blue]模拟运行模式[/]，不会进行任何文件操作。",
        )

        self.description = TextUtils.unwrap(
            """本工具从用户提供的输入中识别[u]预设文件[/]和[u]媒体源文件[/]，
            并基于它们生成一系列任务，并调用 FFmpeg 进行转码。"""
        )

        self.epilog = (
            "[link https://github.com/LambdaXIII/cx-studio-tk]Cxalio Studio Tools[/]"
        )

    @staticmethod
    def show_help(console: r.Console):
        console.print(MKHelp())

    @staticmethod
    def show_full_help(console: r.Console):
        md = importlib.resources.read_text("media_killer", "help.md")
        content = r.Markdown(md, style="default")
        panel = r.Panel(
            content,
            title="Media Killer 教程",
            width=90,
            style="bright_black",
            title_align="left",
        )
        console.print(r.Align.center(panel))



import sys
from collections.abc import Iterable, Generator
from pathlib import Path

from cx_studio.utils import TextUtils
from cx_wealth import IndexedListPanel
from .mission import Mission
from ..appenv import appenv


class ScriptMaker:
    def __init__(self, missions: Iterable[Mission]):
        self.missions = missions
        self._make_dir = "mkdir" if sys.platform == "win32" else "mkdir -p"
        self._default_suffix = ".ps1" if sys.platform == "win32" else ".sh"

    def iter_lines(self) -> Generator[str]:
        folders = set()
        for mission in self.missions:
            output_folders = {
                x.parent.resolve() for x in mission.iter_output_filenames()
            }
            for output_folder in output_folders:
                if not output_folder.exists():
                    yield f"{self._make_dir} {TextUtils.auto_quote(str(output_folder.resolve()))}"
                    folders.add(output_folder)

            es = [mission.ffmpeg]
            es.extend(mission.iter_arguments(quote_mode="auto"))
            yield " ".join(es)

    def save(self, filename: str | Path):
        filename = Path(filename)
        if filename.suffix == "":
            filename = filename.with_suffix(self._default_suffix)

        appenv.check_overwritable_file(filename)

        lines = []
        with open(filename, "w", encoding="utf-8") as f:
            for line in self.iter_lines():
                lines.append(line)
                f.write(line)
                f.write("\n")

        appenv.whisper(IndexedListPanel(lines, title=filename.name, max_lines=999))
        appenv.say("已保存脚本到：{}".format(filename))



from cx_studio.utils import TextUtils
from cx_wealth import WealthHelp


class MKHelp(WealthHelp):
    def __init__(self, **kwargs):
        super().__init__(prog="ffpretty", **kwargs)

        self.description = TextUtils.unwrap(
            """本工具直接转发 FFmpeg 的选项，并在后台调用，
            同时在前台输出美化过的转码信息。
            """
        )

        self.epilog = (
            "[link https://github.com/LambdaXIII/cx-studio-tk]Cxalio Studio Tools[/]"
        )

        trans_group = self.add_group("转码模式", "直接使用ffmpeg的选项即可开始转码。")
        probe_group = self.add_group(
            "查询模式", "直接输入多个媒体文件，查询文件的信息而不进行转码。"
        )



from PIL.Image import Image

from .image_filter import IImageFilter


class ResizeFilter(IImageFilter):
    def __init__(self, width: int, height: int):
        super().__init__()
        self.width = width
        self.height = height

    def run(self, image: Image) -> Image:
        w = image.width if image.width > 0 else image.width
        h = image.height if image.height > 0 else image.height
        return image.resize((w, h))


class FactorResizeFilter(IImageFilter):
    def __init__(self, w_factor: float, h_factor: float):
        super().__init__()
        self.w_factor = w_factor
        self.h_factor = h_factor

    def run(self, image: Image) -> Image:
        wf = self.w_factor if self.w_factor > 0 else 1
        wh = self.h_factor if self.h_factor > 0 else 1
        w = image.width * wf
        h = image.height * wh
        return image.resize((w, h))



from collections.abc import Iterable
from concurrent.futures import ThreadPoolExecutor
from pathlib import Path
from threading import Condition
from time import sleep

from PIL import Image

from cx_studio.utils import PathUtils
from cx_tools.app import SafeError
from cx_wealth import WealthLabel
from cx_wealth import rich_types as r
from .errors import NoSourceFileError, TargetingSourceFileError
from .mission import Mission
from ..appenv import appenv


class MissionRunner:
    def __init__(self, missions: Iterable[Mission], max_workers: int = 10):
        self.missions = list(missions)
        self.max_workers = max_workers
        self.dir_condition = Condition()

    def check_parent(self, target: Path):
        parent = target.parent
        if parent.exists():
            return
        with self.dir_condition:
            if parent.exists():
                return
            appenv.say(f"[yellow]创建目录 {parent}[/]")
            parent.mkdir(parents=True, exist_ok=True)

    def run_mission(self, mission: Mission):
        result_tag = "[green]DONE[/]"
        try:

            if not mission.source.exists():
                raise NoSourceFileError(f"源文件 {mission.source} 不存在")

            target = mission.target
            if target.exists():
                if target == mission.source:
                    raise TargetingSourceFileError(f"目标文件 {target} 与源文件相同")
                if not appenv.context.overwrite:
                    target = PathUtils.ensure_new_file(target)
                    appenv.whisper(
                        f"[yellow]目标文件已存在，已自动重命名为{target.name}。[/]"
                    )

            self.check_parent(target)

            img = Image.open(mission.source)
            img = mission.filter_chain.run(img)
            img.save(target, format=mission.target_format, **mission.saving_options)
        except Image.UnidentifiedImageError:
            appenv.say(f"[red]文件 {mission.source} 无法识别，任务跳过！[/]")
            result_tag = "[red]ERROR[/]"
        except SafeError as e:
            appenv.say(e.message, style=e.style)
            result_tag = "[yellow]SKIPPED[/]"
        except Exception as e:
            appenv.say(f"[red]文件 {mission.source} 处理失败！[/]")
            appenv.say(e)
            result_tag = "[red]UNKOWN ERROR[/]"
        finally:
            appenv.say(r.Columns([WealthLabel(mission), result_tag], expand=True))

    def run(self):
        with appenv.console.status("正在执行任务...") as status:
            with ThreadPoolExecutor(max_workers=self.max_workers) as executor:
                tasks = {
                    m.mission_id: executor.submit(self.run_mission, m)
                    for m in self.missions
                }
                while True:
                    done = [task for task in tasks.values() if task.done()]
                    remains = len(tasks) - len(done)
                    if remains == 0:
                        break
                    status.update(f"正在执行{remains}个任务...")
                    sleep(0.05)



from collections import namedtuple
import sys, os, io
from pathlib import Path, PurePath
from cx_studio.core import FileSize
from chardet import UniversalDetector
from cachetools import LRUCache
from collections.abc import Iterable


class InspectorInfo:
    _sample_cache = LRUCache(maxsize=128)
    _Record = namedtuple("_Record", ["encoding", "sample"])

    def __init__(self, path: os.PathLike, sample_size_hint: int = 2048):
        self.path = Path(path)
        self.size = FileSize(os.path.getsize(path))
        self.encoding: str = "locale"
        self.sample: bytes = b""

        if not self.path.is_file():
            return

        _record: InspectorInfo._Record | None = self._sample_cache.get(path)

        if (
            _record is None
            or len(_record.sample) < sample_size_hint < self.size.total_bytes
        ):
            detector = UniversalDetector()
            detector.reset()
            _sample = bytearray()
            with open(path, "rb") as f:
                while not (detector.done or len(_sample) >= sample_size_hint):
                    raw = f.read(io.DEFAULT_BUFFER_SIZE)
                    if not raw:
                        break
                    detector.feed(raw)
                    _sample.extend(raw)
                if len(_sample) < sample_size_hint:
                    extra = f.read(sample_size_hint - len(_sample))
                    _sample.extend(extra)
                    detector.feed(extra)
                detector.close()
                _encoding = detector.result["encoding"] or "locale"
            _record = InspectorInfo._Record(_encoding, bytes(_sample))
            self._sample_cache[path] = _record

        self.encoding, self.sample = _record.encoding, _record.sample

    def read_lines(
        self, lines: int = -1, ignore_empty_lines: bool = True
    ) -> Iterable[str]:
        with open(self.path, "r", encoding=self.encoding) as fp:
            n = 0
            for line in fp:
                if ignore_empty_lines and not line.strip():
                    continue
                yield line
                n += 1
                if 0 > n >= lines:
                    break

    def peek_first_line(self) -> str:
        for x in self.sample.splitlines(keepends=False):
            if x.strip():
                return x.decode(self.encoding)
        return ""

    def peek_lines(self, ignore_empty_lines: bool = True) -> Iterable[str]:
        for x in self.sample.splitlines(keepends=False):
            if ignore_empty_lines and not x.strip():
                continue
            if x.strip():
                yield x.decode(self.encoding)

    def is_decodable(self) -> bool:
        try:
            self.sample.decode(self.encoding)
            return True
        except UnicodeDecodeError:
            return False



import asyncio
from collections.abc import Iterable
from dataclasses import dataclass
from datetime import datetime

from rich.progress import TaskID

from cx_studio.ffmpeg import FFmpegAsync
from cx_studio.utils.tools import AsyncCanceller
from cx_studio.utils.tools.job_counter import JobCounter
from .mission import Mission
from .mission_runner import MissionRunner, MissionPretender
from ..appenv import appenv

from pathlib import Path


class PoisonError(Exception):
    pass


class MissionMaster:
    @dataclass
    class MInfo:
        mission: Mission
        task_id: TaskID
        total: float | None = None
        runner: MissionRunner | None = None

    def __init__(self, missions: Iterable[Mission], max_workers: int | None = None):
        self._missions = list(missions)
        self._max_workers = max_workers or 1
        self._mission_infos: dict[int, MissionMaster.MInfo] = {}
        self._semaphore = asyncio.Semaphore(self._max_workers)
        self._info_lock = asyncio.Lock()
        self._running_cond = asyncio.Condition()
        self._total_task = appenv.progress.add_task("总进度")

        self._cancel_one = AsyncCanceller()
        self._cancel_all_event = asyncio.Event()

    async def _build_mission_info(self, index):
        mission = self._missions[index]
        ffmpeg = FFmpegAsync(mission.ffmpeg)
        basic_info = await ffmpeg.get_basic_info(mission.source)
        duration = basic_info.get("duration")
        mission_info = MissionMaster.MInfo(
            mission=mission,
            task_id=appenv.progress.add_task(
                mission.name, total=None, visible=False, start=False
            ),
            total=duration.total_seconds if duration else None,
        )

        async with self._info_lock:
            self._mission_infos[index + 1] = mission_info

        if appenv.context.pretending_mode:
            await asyncio.sleep(0.1)

    async def _run_mission(self, index: int):
        async with self._semaphore:
            if self._cancel_all_event.is_set():
                return

            mission_info = self._mission_infos[index]
            mission = mission_info.mission
            runner = (
                MissionPretender(mission)
                if appenv.context.pretending_mode
                else MissionRunner(mission)
            )

            # 记录即将处理的文件列表
            appenv.input_filesize_counter.add_paths(mission.iter_input_filenames())

            async with self._info_lock:
                self._mission_infos[index].runner = runner

            t = asyncio.create_task(runner.execute())

            appenv.progress.start_task(mission_info.task_id)
            try:
                while not t.done():
                    wanna_quit = await self._cancel_one.is_cancelling_async()
                    if wanna_quit or self._cancel_all_event.is_set():
                        runner.cancel()
                        break
                    await asyncio.sleep(0.1)

                # await t
                # 记录已处理完成的文件列表
                appenv.output_filesize_counter.add_paths(
                    mission.iter_output_filenames()
                )

            except asyncio.CancelledError:
                runner.cancel()
                # raise
            finally:
                if appenv.context.pretending_mode:
                    await asyncio.sleep(0.2)
                appenv.progress.stop_task(mission_info.task_id)

    async def _update_tasks(self):
        mission_count = len(self._missions)
        total_time = completed_time = 0
        start_time = datetime.now()

        async with self._info_lock:
            infos = self._mission_infos.copy()

        jobs = JobCounter(mission_count)

        for index, info in infos.items():
            jobs.current = index
            total_time += info.total or 1

            if not info.runner:
                continue

            runner_start_time = info.runner.task_start_time
            if runner_start_time is not None and runner_start_time < start_time:
                start_time = runner_start_time

            desc_str = "[bright_black][{}][{:.2f}x][/][yellow]{}[/]".format(
                jobs.format(), info.runner.task_speed, info.runner.task_description
            )
            if info.runner.is_running():

                appenv.progress.update(
                    info.task_id,
                    visible=True,
                    description=desc_str,
                    completed=info.runner.task_completed,
                    total=info.runner.task_total,
                )

                completed_time += info.runner.task_completed
            else:
                appenv.progress.update(info.task_id, visible=False)
                if info.runner.done():
                    completed_time += info.runner.task_total or 1
        # for
        speed = completed_time / (datetime.now() - start_time).total_seconds()
        desc_str = "[bright_black][{:.2f}x][/][blue]总体进度[/]".format(speed)

        appenv.progress.update(
            self._total_task,
            completed=completed_time,
            total=total_time,
            description=desc_str,
        )

    @staticmethod
    async def _poison_task():
        raise PoisonError()

    async def run(self):
        try:
            self._cancel_all_event.clear()
            # total_start_time = datetime.now()
            async with self._running_cond:
                appenv.progress.update(
                    self._total_task,
                    start=True,
                    visible=True,
                    total=len(self._missions),
                )

                for index, mission in appenv.progress.track(
                    enumerate(self._missions), task_id=self._total_task
                ):
                    appenv.progress.update(self._total_task, description=mission.name)
                    await self._build_mission_info(index)

                workers = [
                    asyncio.create_task(self._run_mission(x))
                    for x in self._mission_infos
                ]

                while not all(x.done() for x in workers):
                    if appenv.wanna_quit_event.is_set():
                        self._cancel_one.cancel()
                        appenv.wanna_quit_event.clear()

                    if appenv.really_wanna_quit_event.is_set():
                        self._cancel_all_event.set()
                        for t in workers:
                            t.cancel()
                        done, pending = await asyncio.wait(
                            workers, return_when=asyncio.ALL_COMPLETED
                        )
                        for p in pending:
                            p.cancel()
                        await asyncio.sleep(0.1)
                        break

                    await self._update_tasks()
                    await asyncio.sleep(0.1)
                # while checking

                await asyncio.gather(*workers, return_exceptions=True)

                # taskgroup
            # running Condition
        except* PoisonError:
            appenv.say("剩余任务被取消")



import sys
from collections.abc import Sequence

from cx_tools.app import IApplication
from cx_wealth import WealthDetailPanel, WealthLabel, IndexedListPanel
from .appenv import appenv
from .components.mission_runner import MissionRunner
from .simple_appcontext import SimpleAppContext, SimpleHelp
from .simple_filter_chain_builder import SimpleFilterChainBuilder
from .simple_mission_builder import SimpleMissionBuilder


class JpeggerApp(IApplication):
    def __init__(self, arguments: Sequence[str] | None = None):
        super().__init__(arguments or sys.argv[1:])

    def start(self):
        appenv.context = SimpleAppContext.from_arguments(self.sys_arguments)
        appenv.start()

    def stop(self):
        appenv.stop()

    def run(self):
        if appenv.context.show_help:
            appenv.say(SimpleHelp())
            return

        appenv.whisper(WealthDetailPanel(appenv.context, title="初始化参数"))

        filter_chain = SimpleFilterChainBuilder.build_filter_chain_from_simple_context(
            appenv.context
        )
        appenv.whisper(WealthDetailPanel(filter_chain, title="过滤器链"))

        if not appenv.context.inputs:
            appenv.say("未指定输入文件，无事可做")
            return

        builder = SimpleMissionBuilder(filter_chain, appenv.context)

        missions = builder.make_missions(appenv.context.inputs)

        appenv.whisper(
            IndexedListPanel([WealthLabel(x) for x in missions], title="任务列表")
        )

        runner = MissionRunner(missions)
        runner.run()



from .simple_application import JpeggerApp


def run():
    from rich.traceback import install

    install(show_locals=False, word_wrap=True, suppress=["rich"])
    with JpeggerApp() as app:
        app.run()



import tomllib
# from dataclasses import dataclass, field
from pathlib import Path

from box import Box
from pydantic import BaseModel, Field, ConfigDict
from rich.columns import Columns

from cx_studio.utils import PathUtils, TextUtils

DefaultSuffixes = (
    ".mov .mp4 .mkv .avi .wmv .flv .webm "
    ".m4v .ts .m2ts .m2t .mts .m2v .m4v "
    ".vob .3gp .3g2 .f4v .ogv .ogg .mpg "
    ".mpeg .mxf .asf .rm .rmvb .divx "
    ".xvid .h264 .h265 .hevc .vp8 "
    ".vp9 .av1 .avc .avchd .flac .mp3 .wav "
    ".m4a .aac .ogg .wma .flac .alac .aiff "
    ".ape .dsd .pcm .ac3 .dts .eac3 .mp2 "
    ".mpa .opus .mka .mkv .webm .flv .ts .m2ts "
    ".m2t .mts .m2v .m4v .vob .wav .m4a .aac "
    ".ogg .wma .flac .aiff .ape .dsd .pcm "
    ".ac3 .dts .eac3 .mp2 .mpa .opus .mka .mxf_op1a"
)


# @dataclass(frozen=True)
class Preset(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, frozen=True)

    id: str = ""
    name: str = ""
    description: str = ""
    path: Path = Path("")
    ffmpeg: str = "ffmpeg"
    overwrite: bool = False
    hardware_accelerate: str | None = "auto"
    options: str | list = ""
    source_suffixes: set = Field(default_factory=set)
    target_suffix: str = ""
    target_folder: Path = Path(".")
    keep_parent_level: int = 0
    inputs: list = Field(default_factory=list)
    outputs: list = Field(default_factory=list)
    custom: dict = Field(default_factory=dict)
    # raw: DataPackage = Field(default_factory=DataPackage)
    raw: Box = Box()

    @staticmethod
    def _get_source_suffixes(data: Box) -> set[str]:
        default_suffixes = (
            set(DefaultSuffixes.split())
            if not data.source.ignore_default_suffixes  # type:ignore
            else set()
        )
        includes = {
            PathUtils.normalize_suffix(s)
            for s in TextUtils.auto_list(data.source.suffix_includes)  # type:ignore
        }
        excludes = {
            PathUtils.normalize_suffix(s)
            for s in TextUtils.auto_list(data.source.suffix_excludes)  # type:ignore
        }
        return default_suffixes | includes - excludes

    @classmethod
    def load(cls, filename: Path | str):
        filename = PathUtils.force_suffix(filename, ".toml")
        with open(filename, "rb") as f:
            toml = tomllib.load(f)
        # data = DataPackage(**toml)
        data = Box(toml)

        return cls(
            id=data.general.preset_id,  # type:ignore
            name=data.general.name,  # type:ignore
            description=data.general.description,  # type:ignore
            path=Path(filename).resolve(),
            ffmpeg=data.general.ffmpeg,  # type:ignore
            overwrite=data.general.overwrite,  # type:ignore
            hardware_accelerate=data.general.hardware_accelerate,  # type:ignore
            options=data.general.options,  # type:ignore
            source_suffixes=Preset._get_source_suffixes(data),
            target_suffix=data.target.suffix,  # type:ignore
            target_folder=data.target.folder,  # type:ignore
            keep_parent_level=data.target.keep_parent_level,  # type:ignore
            inputs=data.input,  # type:ignore
            outputs=data.output,  # type:ignore
            custom=data.custom.to_dict(),  # type:ignore
            raw=data,
        )

    def __eq__(self, value: object) -> bool:
        if not isinstance(value, Preset):
            return False
        return self.id == value.id and self.path == value.path

    def __hash__(self) -> int:
        return hash(self.id) ^ hash(self.path) ^ hash("preset")

    def __rich_detail__(self):
        yield "ID", self.id
        yield "预设名称", self.name
        yield "预设描述", self.description
        yield "预设文件路径", str(self.path)
        yield "FFmpeg路径", self.ffmpeg
        yield "是否覆盖", str(self.overwrite)
        yield "硬件加速模式", self.hardware_accelerate
        yield "额外参数", self.options
        yield "源文件扩展名", Columns(self.source_suffixes)
        yield "目标文件扩展名", self.target_suffix
        yield "目标文件夹", str(self.target_folder)
        yield "保留父级层级", str(self.keep_parent_level)
        yield "输入参数", self.inputs
        yield "输出参数", self.outputs
        yield "自定义参数", self.custom
        yield "原始数据", self.raw

    def __rich_label__(self):
        yield "[bold bright_black]P[/]"
        yield f"[green][{len(self.inputs)}->{len(self.outputs)}][/green]"
        yield f"[cyan]{self.name}[/cyan]"
        yield f"[italic yellow]{self.description}[/italic yellow]"
        yield f"[bright_black]({self.path})[/bright_black]"



import os
from pathlib import PurePath, Path
from typing import Self
from collections.abc import Iterable
from media_scout.inspectors.media_path_inspector import MediaPathInspector
from .edl_inspector import EDLInspector
from .fcpxml_inspector import FCPXMLInspector, FCPXMLDInspector
from .inspector_info import InspectorInfo
from .legacy_xml_inspector import LegacyXMLInspector
from .resolve_metadata_inspector import ResolveMetadataInspector


class InspectorChain:
    def __init__(
        self,
        *inspectors: MediaPathInspector,
        allow_duplicated: bool = False,
        auto_resolve: bool = False,
    ):
        self.allow_duplicated = allow_duplicated
        self.auto_resolve = auto_resolve
        self._inspectors: list[MediaPathInspector] = []
        self._inspectors.extend(inspectors)
        self._exported_paths: set[PurePath] = set()

    def add_inspector(self, inspector: MediaPathInspector) -> Self:
        self._inspectors.append(inspector)
        return self

    def _auto_resolve(self, path: os.PathLike, info: InspectorInfo) -> PurePath:
        x = Path(path)
        if not self.auto_resolve:
            return x
        if not x.is_absolute():
            x = Path(info.path.parent, x)
        return x.resolve() if x.exists() else x

    def inspect(self, info: InspectorInfo) -> Iterable[PurePath]:
        for inspector in self._inspectors:
            if inspector.is_inspectable(info):
                for path in inspector.inspect(info):
                    path = self._auto_resolve(path, info)
                    if self.allow_duplicated or path not in self._exported_paths:
                        self._exported_paths.add(path)
                        yield path
                break

    def __enter__(self) -> Self:
        self._exported_paths.clear()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._exported_paths.clear()
        return False



from abc import ABC, abstractmethod
from collections.abc import Sequence
import sys
from .safe_error import SafeError


class IApplication(ABC):
    def __init__(self, arguments: Sequence[str] | None = None):
        self.sys_arguments = arguments or sys.argv[1:]

    @abstractmethod
    def start(self):
        pass

    @abstractmethod
    def stop(self):
        pass

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()
        return False

    @abstractmethod
    def run(self):
        pass



from cx_studio.core.cx_filesize import FileSize
from cx_studio.core.cx_time import CxTime
from pathlib import Path
from typing import Any
import cx_wealth.rich_types as r


class FormatInfo:
    def __init__(self, data: dict[str, Any]):
        self.data = data

    @property
    def format_name(self) -> str:
        return self.data.get("format_name", None)

    @property
    def format_long_name(self) -> str:
        return self.data.get("format_long_name", None)

    @property
    def duration(self) -> CxTime:
        duration_text = self.data.get("duration", None)
        if duration_text:
            return CxTime.from_seconds(float(duration_text))
        return None

    @property
    def bit_rate(self) -> FileSize:
        bit_rate_text = self.data.get("bit_rate", None)
        if bit_rate_text:
            return FileSize.from_bytes(int(bit_rate_text))
        return None

    @property
    def size(self) -> FileSize:
        size_text = self.data.get("size", None)
        if size_text:
            return FileSize.from_bytes(int(size_text))
        return None

    @property
    def streams(self) -> int:
        return self.data.get("nb_streams", 0)

    @property
    def filename(self) -> str:
        name = self.data.get("filename", None)
        if not name:
            return None
        return Path(name).name

    def __rich_detail__(self):
        name = self.filename
        if name:
            yield "文件名", r.Text(name, style="red")

        format_name = self.format_name
        if format_name:
            yield "编码代码", r.Text(format_name, style="green")

        format_long_name = self.format_long_name
        if format_long_name:
            yield "混流格式", r.Text(format_long_name, style="blue")

        duration = self.duration
        if duration:
            yield "时长", r.Text(duration.pretty_string, style="cyan")

        bit_rate = self.bit_rate
        if bit_rate:
            yield "码率", r.Text(f"{bit_rate.pretty_string}/s", style="yellow")

        size = self.size
        if size:
            yield "大小", r.Text(size.pretty_string, style="yellow")



from collections.abc import Generator, Iterable
from pathlib import Path, PurePath

from cx_studio.path_expander import PathExpander, SuffixValidator
from media_killer.appenv import appenv
from media_scout.inspectors import (
    EDLInspector,
    FCPXMLDInspector,
    FCPXMLInspector,
    FileListInspector,
    InspectorChain,
    ResolveMetadataInspector,
)
from media_scout.inspectors.inspector_info import InspectorInfo
from .preset import Preset


class SourceExpander:
    def __init__(self, preset: Preset):
        self.preset = preset
        self._exported_paths = set()
        self._source_inspector = InspectorChain(
            EDLInspector(),
            FCPXMLDInspector(),
            FCPXMLInspector(),
            ResolveMetadataInspector(),
            FileListInspector(".txt", ".csv"),
            auto_resolve=True,
        )

    def _pre_expand(self, *paths) -> Iterable[PurePath]:
        for path in paths:
            path = Path(path)
            info = InspectorInfo(path)
            inspected = list(self._source_inspector.inspect(info))
            if inspected:
                yield from inspected
            else:
                yield path

    def expand(self, *paths) -> Generator[Path]:
        expander_start_info = PathExpander.StartInfo(
            accept_dirs=False,
            file_validator=SuffixValidator(self.preset.source_suffixes),
        )

        expander = PathExpander(expander_start_info)
        for source in self._pre_expand(*paths):
            wanna_quit = False
            if appenv.wanna_quit_event.is_set():
                wanna_quit = True
                appenv.wanna_quit_event.clear()
            if wanna_quit:
                appenv.whisper("接收到[bold]取消信号[/bold]，中断路径展开操作。")
                break
            for p in expander.expand(Path(source)):
                if p in self._exported_paths:
                    continue
                self._exported_paths.add(p)
                yield p



import os
from pathlib import Path, PurePath
import time
from cx_tools.app import IApplication
import sys
from collections.abc import Iterable
from .appenv import appenv
from media_scout.inspectors.filelist_inspector import FileListInspector
from cx_studio.utils import PathUtils, TextUtils

from cx_wealth import WealthDetailPanel
from rich.rule import Rule

from .inspectors import (
    ResolveMetadataInspector,
    MediaPathInspector,
    InspectorInfo,
    EDLInspector,
    LegacyXMLInspector,
    FCPXMLInspector,
    FCPXMLDInspector,
    InspectorChain,
)
from .arg_parser import MSHelp


class Application(IApplication):
    APP_NAME = "MediaScout"
    APP_VERSION = "0.1.0"

    def __init__(self):
        super().__init__()

    def start(self):
        appenv.start()
        appenv.show_banner()
        appenv.whisper("MediaScout 启动")
        appenv.whisper(WealthDetailPanel(appenv.context))

    def stop(self):
        appenv.stop()
        appenv.whisper("Bye~")

    @staticmethod
    def resolve(path: os.PathLike) -> str | None:
        result = Path(path)
        if appenv.context.existed_only and not result.exists():
            appenv.whisper("[red]{} 不存在[/]".format(result))
            return None
        if appenv.context.auto_resolve:
            result = result.resolve()
        return PathUtils.quote(result, appenv.context.quote_mode)

    @staticmethod
    def auto_expand(path: os.PathLike, info: InspectorInfo) -> Iterable[PurePath]:
        result = Path(path)
        includes = [info.path.parent.resolve()] if appenv.context.auto_resolve else []
        includes.extend([Path(x) for x in appenv.context.includes])

        if result.is_absolute() or not appenv.context.includes:
            yield result
        else:
            appenv.whisper("[red]在搜索路径中搜索：{}[/]".format(result))
            for include in includes:
                p = Path(include).absolute() / result
                if p.exists():
                    appenv.whisper("找到：{}".format(p))
                    yield p

    def iter_results(self):
        inspectors = [
            ResolveMetadataInspector(),
            EDLInspector(),
            LegacyXMLInspector(),
            FCPXMLInspector(),
            FCPXMLDInspector(),
            FileListInspector(".txt", ".ps1", ".sh"),
        ]

        chain = InspectorChain(
            *inspectors, allow_duplicated=appenv.context.allow_duplicated
        )

        for path in appenv.context.inputs:
            path = Path(path)
            appenv.say(Rule(path.name, style="dim green"))
            info = InspectorInfo(Path(path))
            for result in chain.inspect(info):
                for x in self.auto_expand(result, info):
                    if a := self.resolve(x):
                        yield a

    def run(self):
        if appenv.context.show_help:
            MSHelp.show_help(appenv.console)
            return

        if appenv.context.show_full_help:
            MSHelp.show_full_help(appenv.console)
            return

        if appenv.context.allow_duplicated:
            appenv.say("[red]允许输出重复项[/]")
            time.sleep(0.5)
        if appenv.context.auto_resolve:
            appenv.say("[yellow]自动整理或折叠路径[/]")
            time.sleep(0.5)
        if appenv.context.existed_only:
            appenv.say("[green]只输出存在的文件[/]")
            time.sleep(0.5)

        result = []
        for x in self.iter_results():
            result.append(x)
            appenv.print(x)

        appenv.say("[yellow]共找到 {} 个媒体路径。[/]".format(len(result)))

        if appenv.context.output:
            output_file = PathUtils.auto_suffix(appenv.context.output, ".txt")
            with open(output_file, "w") as fp:
                for x in result:
                    fp.write(str(x) + "\n")

            appenv.say('[green]列表已保存到："{}"[/]'.format(output_file))



class SafeError(Exception):
    def __init__(self, message: str | None = None, style: str | None = None):
        super().__init__(message)
        self.message = message
        self.style = style or "cx.error"



import datetime
from pathlib import Path

from cx_studio.utils import TextUtils


class ConfigManager:
    def __init__(self, app_name: str):
        self.app_name = app_name.strip()

    @property
    def config_dir(self) -> Path:
        return Path.home() / ".config" / "cx-studio" / self.app_name

    @property
    def log_dir(self) -> Path:
        return self.config_dir / "logs"

    def new_log_file(self):
        filename = f"{datetime.datetime.now().strftime("%Y%m%d_%H%M%S")}_{TextUtils.random_string(5)}.log"
        self.log_dir.mkdir(parents=True, exist_ok=True)
        return self.log_dir / filename

    def remove_old_log_files(self, keep=10):
        log_files = list(self.log_dir.glob("*.log"))
        log_files.sort(key=lambda x: x.stat().st_mtime)
        if len(log_files) > keep:
            for file in log_files[:-keep]:
                file.unlink()

    def get_file(self, filename: str) -> Path:
        return self.config_dir / filename



from .format_info import FormatInfo
from .stream_info import StreamInfo
import cx_wealth.rich_types as r
from cx_wealth import WealthDetailTable, WealthLabel


class MediaInfo:
    def __init__(self, data: dict):
        self.data = data
        self.format_info = FormatInfo(data.get("format", {}))
        self.stream_infos = [StreamInfo(stream) for stream in data.get("streams", [])]

    def __rich_console__(self, console, options):
        o = options.update(highlight=False)
        table = WealthDetailTable(self.format_info, sub_box=False)

        labels = r.Group(*(WealthLabel(stream) for stream in self.stream_infos))
        stream_box = r.Panel(
            labels,
            title="流信息",
            title_align="left",
            border_style="dim",
        )

        group = r.Group(table, stream_box)
        box = r.Panel(
            group,
            title="媒体信息",
            title_align="left",
            subtitle=self.format_info.filename,
            subtitle_align="right",
        )
        return console.render(box, options=o)



from cx_tools.app import IApplication
from collections.abc import Sequence
from .appenv import appenv


class LandlordApp(IApplication):
    def __init__(self, arguments: Sequence[str] | None = None):
        super().__init__(arguments)

    def start(self):
        appenv.start()

    def stop(self):
        appenv.stop()

    def run(self):
        appenv.say("Hello, World!")



from .application import FFPrettyApp


def run() -> int:
    from rich.traceback import install

    install(show_locals=False, word_wrap=True, suppress=["rich"])

    with FFPrettyApp() as app:
        result = app.run()
        return 0 if result else -1



from collections.abc import Iterable
from .media_path_inspector import MediaPathInspector, InspectorInfo
from pathlib import PurePath, Path
import re
import xml.etree.ElementTree as ET
import urllib.parse


class FCPXMLInspector(MediaPathInspector):
    DOCTYPE_PATTERN = re.compile(r"<!DOCTYPE\s+fcpxml>")
    URL_HEAD = re.compile(r"^file://.*?/")

    def _is_inspectable(self, info: InspectorInfo) -> bool:
        if info.path.suffix != ".fcpxml":
            return False
        if not info.is_decodable():
            return False
        for x in info.peek_lines():
            if self.DOCTYPE_PATTERN.search(x):
                return True
        return False

    def _inspect(self, info: InspectorInfo) -> Iterable[PurePath]:
        tree = ET.parse(info.path)
        root = tree.getroot()
        for node in root.iter("media-rep"):
            kind = node.get("kind")
            if kind == "original-media":
                url = urllib.parse.unquote(node.get("src") or "")
                if self.URL_HEAD.match(url):
                    yield PurePath(self.URL_HEAD.sub("", url))


class FCPXMLDInspector(MediaPathInspector):
    def __init__(self) -> None:
        super().__init__()
        self._inspector = FCPXMLInspector()

    def _is_inspectable(self, info: InspectorInfo) -> bool:
        if info.path.suffix != ".fcpxmld":
            return False
        if not info.path.is_dir():
            return False
        core_file = Path(info.path, "Info.fcpxml")
        core_info = InspectorInfo(core_file)
        return self._inspector.is_inspectable(core_info)

    def _inspect(self, info: InspectorInfo) -> Iterable[PurePath]:
        core_file = Path(info.path, "Info.fcpxml")
        core_info = InspectorInfo(core_file)
        yield from self._inspector.inspect(core_info)



import asyncio
import itertools
import os
import random
from datetime import datetime
from pathlib import Path
from pprint import saferepr
from typing import override

from cx_studio.core.cx_time import CxTime
from cx_studio.ffmpeg import FFmpegAsync
from cx_studio.utils import PathUtils
from cx_wealth import rich_types as r
from cx_wealth.indexed_list_panel import IndexedListPanel
from cx_wealth.wealth_detail import WealthDetailPanel
from media_killer.appenv import appenv
from .exception import SafeError
from .mission import Mission


class MissionRunner:
    def __init__(self, mission: Mission):
        self.mission = mission
        self._ffmpeg: FFmpegAsync = FFmpegAsync(self.mission.ffmpeg)
        self._input_files = [self.mission.source] + list(
            self.mission.iter_input_filenames()
        )
        self._output_files = [self.mission.standard_target] + list(
            self.mission.iter_output_filenames()
        )

        self._task_description: str = self.mission.name
        self._task_completed: float = 0
        self._task_total: float | None = None
        self._task_speed: float = 0

        self._cancel_event = asyncio.Event()

        self._start_time: datetime | None = None
        self._end_time: datetime | None = None
        self._running_cond = asyncio.Condition()
        self._ffmpeg_outputs = []

    def cancel(self):
        # self._canceled = True
        self._cancel_event.set()

    @property
    def task_description(self):
        return self._task_description

    @property
    def task_completed(self):
        return self._task_completed

    @property
    def task_total(self):
        return self._task_total

    @property
    def task_start_time(self):
        return self._start_time

    @property
    def task_end_time(self):
        return self._end_time

    @property
    def task_speed(self):
        return self._task_speed

    def done(self):
        return self._start_time is not None and self._end_time is not None

    def is_running(self):
        return self._running_cond.locked()

    def make_line_report(self, right_side: str):
        header = "[bright_black]M[/] [dim green][{i_count}->{o_count}][/] ".format(
            i_count=len(self.mission.inputs), o_count=len(self.mission.outputs)
        )
        name = "[yellow]{}[/]".format(self.mission.name)

        label = header + name

        left = r.Text.from_markup(label, end="", justify="left", overflow="ellipsis")
        left.no_wrap = True
        right = r.Text.from_markup(right_side, justify="right")

        return r.Columns([left, right], expand=True)

    async def _on_started(self):
        report = self.make_line_report("[yellow]开始[/]")
        g = r.Group(
            WealthDetailPanel(self.mission, title=str(self.mission.mission_id)), report
        )
        appenv.whisper(g)

    async def _on_progress_updated(self, c: CxTime, t: CxTime | None):
        c_seconds = c.total_seconds
        t_seconds = t.total_seconds if t else None
        self._task_completed = c_seconds
        self._task_total = t_seconds
        self._task_speed = self._ffmpeg.coding_info.current_speed

    async def _on_finished(self):
        appenv.say(self.make_line_report("[green]完成[/]"))

    async def _on_terminated(self):
        appenv.whisper(IndexedListPanel(self._ffmpeg_outputs, title="FFmpeg 输出"))
        appenv.say(self.make_line_report("[red]运行异常[/]"))
        await self._clean_up()

    async def _on_canceled(self, reason: str | None = None):
        appenv.say(
            self.make_line_report("[bright_blue]{}[/]".format(reason or "被取消"))
        )
        await self._clean_up()
        if self._cancel_event.is_set():
            self._cancel_event.clear()

    async def _clean_up(self):
        self._ffmpeg_outputs.clear()
        safe_outputs = set(self._output_files) - set(self._input_files)
        deleting_files = set(filter(lambda x: x.exists(), safe_outputs))
        if deleting_files:
            appenv.whisper(IndexedListPanel(deleting_files, title="未完成的目标文件"))
            appenv.add_garbage_files(*deleting_files)

    async def _on_verbose(self, line: str):
        self._ffmpeg_outputs.append(line)

    def _prepare_mission(self):
        conflicts = set(self._input_files) & set(self._output_files)
        if len(conflicts) > 0:
            appenv.whisper(IndexedListPanel(conflicts, title="发现重叠文件"))
            raise SafeError("检测到重叠的输入输出文件")

        if not PathUtils.is_executable(Path(self._ffmpeg.executable)):
            raise SafeError("ffmpeg可执行文件无效:{}".format(self._ffmpeg.executable))

        no_existed_input_files = set(
            itertools.filterfalse(lambda a: a.exists(), self._input_files)
        )
        if no_existed_input_files:
            raise SafeError(
                "输入文件不存在: {}".format(";".join(map(str, no_existed_input_files)))
            )

        o_dirs = set(map(lambda a: a.parent, self._output_files))
        invalid_o_dirs = set(
            itertools.filterfalse(
                lambda a: os.access(a, os.W_OK),
                filter(lambda a: a.exists(), o_dirs),
            )
        )
        if invalid_o_dirs:
            raise SafeError("输出目录无效")

        non_existent_o_dirs = set(itertools.filterfalse(lambda a: a.exists(), o_dirs))
        if non_existent_o_dirs:
            self._task_description = "创建目标文件夹"
            for x in non_existent_o_dirs:
                x.mkdir(parents=True, exist_ok=True)
            appenv.whisper(
                IndexedListPanel(non_existent_o_dirs, title="自动创建目标文件夹")
            )
        self._task_description = self.mission.name

    async def execute(self):
        async with self._running_cond:
            self._ffmpeg.add_listener("started", self._on_started)
            self._ffmpeg.add_listener("progress_updated", self._on_progress_updated)
            self._ffmpeg.add_listener("finished", self._on_finished)
            self._ffmpeg.add_listener("canceled", self._on_canceled)
            self._ffmpeg.add_listener("terminated", self._on_terminated)
            self._ffmpeg.add_listener("verbose", self._on_verbose)

            self._start_time = datetime.now()
            self._task_description = self.mission.name

            result = None

            try:
                self._prepare_mission()

                main_task = asyncio.create_task(
                    self._ffmpeg.execute(self.mission.iter_arguments())
                )

                while not main_task.done():
                    await asyncio.sleep(0.1)
                    if self._cancel_event.is_set():
                        self._ffmpeg.cancel()
                        self._cancel_event.clear()
                        break

                # await asyncio.wait([main_task])
                result = main_task.result()

            except asyncio.CancelledError:
                self.cancel()
                result = False

            except SafeError as e:
                await self._on_canceled(reason=e.message)
                result = False

            finally:
                # async with self._cancelling_cond:
                # await asyncio.wait([main_task])
                self._end_time = datetime.now()
                await self._ffmpeg.wait_for_complete()
                return result

        # running condition


class MissionPretender(MissionRunner):
    def _init__(self, mission: Mission):
        super().__init__(mission)

    async def _pretending_prepare_mission(self):
        self._task_description = "检查输入输出文件"
        await asyncio.sleep(0.5)
        conflicts = set(self._input_files) & set(self._output_files)
        if len(conflicts) > 0:
            appenv.whisper(IndexedListPanel(conflicts, title="发现重叠文件"))
            raise SafeError("检测到重叠的输入输出文件")

        self._task_description = "检查ffmpeg可执行文件"
        await asyncio.sleep(0.3)
        if not PathUtils.is_executable(Path(self._ffmpeg.executable)):
            raise SafeError("ffmpeg可执行文件无效:{}".format(self._ffmpeg.executable))

        self._task_description = "检查输入文件"
        await asyncio.sleep(0.5)
        no_existed_input_files = set(
            itertools.filterfalse(lambda x: x.exists(), self._input_files)
        )
        if no_existed_input_files:
            raise SafeError(
                "输入文件不存在: {}".format(
                    ";".join(map(saferepr, no_existed_input_files))
                )
            )

        self._task_description = "检查输出目录"
        await asyncio.sleep(0.2)
        o_dirs = set(map(lambda x: x.parent, self._output_files))
        invalid_o_dirs = set(
            itertools.filterfalse(
                lambda x: os.access(x, os.W_OK),
                filter(lambda x: x.exists(), o_dirs),
            )
        )
        if invalid_o_dirs:
            appenv.whisper(IndexedListPanel(invalid_o_dirs, title="无效的输出目录"))
            raise SafeError("输出目录无效")

        self._task_description = "创建输出目录"
        non_existent_o_dirs = set(itertools.filterfalse(lambda x: x.exists(), o_dirs))
        if non_existent_o_dirs:
            self._task_total = len(non_existent_o_dirs)
            for _ in non_existent_o_dirs:
                self._task_completed += 1
                await asyncio.sleep(0.25)
            self._task_total = None
            appenv.whisper(
                IndexedListPanel(non_existent_o_dirs, title="自动创建目标文件夹")
            )

    @override
    async def execute(self):
        async with self._running_cond:
            try:
                await self._pretending_prepare_mission()

                self._task_total = 1000
                self._task_completed = 0
                self._task_description = self.mission.name
                self._start_time = datetime.now()

                while self._task_completed < self._task_total:
                    step = random.randint(5, 25)
                    self._task_completed += step
                    self._task_speed = random.randint(1200, 9999) / 100
                    await asyncio.sleep(0.05)
                    if self._cancel_event.is_set():
                        await self._on_canceled()
                        break
            except asyncio.CancelledError:
                self.cancel()
            except SafeError as e:
                await self._on_canceled(reason=e.message)
            finally:
                self._end_time = datetime.now()
                return True



from pathlib import Path
import asyncio
import json

from cx_studio.path_expander import CmdFinder
from cx_studio.utils import AsyncStreamUtils
from cx_tools.app.safe_error import SafeError
from .appenv import appenv
from typing import Any
from .info_elements import MediaInfo


class Prober:
    def __init__(self, ffprobe_executable: str | Path | None = None):
        # 尝试找到ffprobe可执行文件，如果没有指定，则使用默认路径
        self._ffprobe_executable = str(CmdFinder.which(ffprobe_executable or "ffprobe"))

    async def get_details(self, file: Path) -> dict[str, Any]:
        """使用ffprobe获取媒体文件的详细信息"""
        args = [
            self._ffprobe_executable,
            "-v",
            "quiet",
            "-print_format",
            "json",
            "-show_format",
            "-show_streams",
            str(file),
        ]

        process = await AsyncStreamUtils.create_subprocess(
            *args, stdout=asyncio.subprocess.PIPE, stderr=asyncio.subprocess.PIPE
        )

        stdout, stderr = await process.communicate()

        if process.returncode != 0:
            error_msg = stderr.decode("utf-8", errors="ignore").strip()
            raise SafeError(f"ffprobe执行失败: {error_msg}")

        try:
            return json.loads(stdout.decode("utf-8"))
        except json.JSONDecodeError as e:
            raise SafeError(f"解析ffprobe输出失败: {str(e)}")

    def probe(self, file: Path) -> dict[str, Any]:
        """同步获取媒体文件的详细信息"""
        details = asyncio.run(self.get_details(file))
        media_info = MediaInfo(details)
        appenv.say(media_info)
